""" Tests for generic and EVLA CAL manifests """

# TODO when the time comes: tests of ingestion manifests for each remaining ingestion type
#  as they're implemented, in separate modules so they don't get too unwieldy

import json
import logging
import re
import shutil
import sys
from pathlib import Path

# pylint: disable=E0401, E0402, R1721, W0621

import pytest

from ingest_envoy.ingestion_manifest import (
    IngestionManifestBuilder,
    build_manifest_filename,
    find_manifest,
)
from ingest_envoy.manifest_components import (
    MANIFEST_NAME_BASE,
    MANIFEST_NAME_EXT,
    IngestionManifestKey,
    ParamsKey,
    InputScienceProduct,
    InputGroup,
    ManifestParameters,
    OutputScienceProduct,
    AncillaryProduct,
    OutputGroup,
    ARTIFACT_EXT,
    ARTIFACT_NAME,
)
from ingest_envoy.utilities import (
    ScienceProductType,
    AncillaryProductType,
    Telescope,
)

# pylint: disable=E0401, E1120

# ingest_path is NOT unused! Don't let IJ remove the import!
from .conftest import ingest_path, populate_fake_evla_cal_ingest_path, WANTED_FILENAMES, UNWANTED

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

FAKE_LOCATOR = "uid://evla/calibration/doo-wah-ditty-ditty-af123"


def test_manifest_is_complete(ingest_path: Path):
    """
    Most ingestion manifests should have parameters, an input group, and an output group.
    An output group will contain one or more science products, and sometimes ancillary products.

    :return:
    """

    populate_fake_evla_cal_ingest_path(ingest_path)
    assert isinstance(ingest_path, Path)
    params_expected = ManifestParameters(
        telescope=Telescope.EVLA,
        ngas_ingest=False,
        reingest=False,
        calibrate=False,
        staging_source_dir=ingest_path,
    )

    sp1 = InputScienceProduct(
        sp_type=ScienceProductType.EVLA_CAL,
        locator=FAKE_LOCATOR,
    )

    ig_in = InputGroup(science_products=[sp1])
    osp_in = OutputScienceProduct(
        type=ScienceProductType.EVLA_CAL, filename="my_science_products.tar"
    )
    ap_in = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename="weblog.tgz")

    manifest, _ = IngestionManifestBuilder(
        staging_source_dir=ingest_path,
        telescope=Telescope.EVLA,
        sp_type=ScienceProductType.EVLA_CAL,
        locator=FAKE_LOCATOR,
    ).build()

    assert manifest.parameters == params_expected
    assert manifest.input_group == ig_in
    assert manifest.output_group.science_products[0] == osp_in
    assert ap_in in manifest.output_group.ancillary_products

    af_tar_candidates = [
        file
        for file in ingest_path.iterdir()
        if file.name.startswith(ARTIFACT_NAME) and file.name.endswith(ARTIFACT_EXT)
    ]
    assert len(af_tar_candidates) == 1

    shutil.rmtree(ingest_path)


def test_builds_expected_manifest_filename():
    """
    We expect the manifest to be named like

    ingestion_manifest_2019_07_30_T13_03_00.936.json

    :return:
    """
    filename = build_manifest_filename()

    assert filename.startswith(MANIFEST_NAME_BASE)
    assert filename.endswith(MANIFEST_NAME_EXT)

    filename_parts = filename.split("_")
    assert len(filename_parts) == 7

    # get just the timestamp
    timestamp = filename.replace(MANIFEST_NAME_BASE, "").replace(MANIFEST_NAME_EXT, "")

    # we should have gotten year, month, day, hours, minutes, seconds to 3 decimal places
    assert re.match(r"\d{4}_\d{2}_\d{2}T\d{2}_\d{2}_\d{2}\.\d{0,3}", timestamp)


def test_filters_cal_input_files(ingest_path: Path):
    """
    We'll be getting calibration/image/eb, etc. science products from a directory under
    /lustre/aoc/cluster/pipeline/{CAPO_PROFILE}/workspaces/staging

    Make sure we take -only- the files to be ingested.

    :param ingest_path: our temporary dir
    """

    populate_fake_evla_cal_ingest_path(ingest_path)
    locator = "uid://evla/calibration/twinkle-twinkle-little-quasar"
    manifest, _ = IngestionManifestBuilder(
        telescope=Telescope.EVLA,
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.EVLA_CAL,
        locator=locator,
    ).build()

    # does manifest contain all the data it should, and none of the data it shouldn't?

    assert manifest.locator == locator
    params = manifest.parameters
    assert not params.reingest and not params.ngas_ingest and not params.calibrate

    input_group = manifest.input_group
    assert len(input_group.science_products) == 1
    sp_in = input_group.science_products[0]
    assert sp_in.type == ScienceProductType.EVLA_CAL

    output_group = manifest.output_group
    assert len(output_group.science_products) == 1
    assert len(output_group.ancillary_products) == 2
    for product in output_group.ancillary_products:
        if product.filename not in WANTED_FILENAMES:
            assert product.filename.startswith(ARTIFACT_NAME) and product.filename.endswith(
                ARTIFACT_EXT
            )
        assert product.filename not in UNWANTED

    sp_out = output_group.science_products[0]
    assert sp_out.type == ScienceProductType.EVLA_CAL

    assert sp_out.filename in WANTED_FILENAMES
    assert sp_out.filename not in UNWANTED

    shutil.rmtree(ingest_path)


def test_writes_expected_output_files(ingest_path: Path):
    """
    Did the manifest builder produce the manifest file, the weblog, and the science product tar?

    :param ingest_path: ingestion location
    :return:
    """
    populate_fake_evla_cal_ingest_path(ingest_path)
    manifest_file, manifest = IngestionManifestBuilder(
        telescope=Telescope.EVLA,
        staging_source_dir=ingest_path,
        locator="uid://evla/calibration/fee-fi-fo-fum-acdf23",
        sp_type=ScienceProductType.EVLA_CAL,
    ).build()
    assert manifest_file
    assert manifest

    ingestion_files = [file for file in ingest_path.iterdir()]

    # at a minimum, we expect the manifest, the ingestion artifact, and the science products tar
    assert len(ingestion_files) >= 3
    mf_json = [
        file
        for file in ingestion_files
        if file.name.startswith(MANIFEST_NAME_BASE) and file.name.endswith(MANIFEST_NAME_EXT)
    ][0]
    assert mf_json
    tars = [file for file in ingestion_files if file.name.endswith(".tar")]
    assert len(tars) >= 2

    shutil.rmtree(ingest_path)


def test_params_json_well_formed():
    """
    Make sure our ManifestParameters makes nice JSON

    :return:
    """
    telescope = Telescope.EVLA

    params_dict = {
        ParamsKey.PARAMETERS.value: {
            ParamsKey.TELESCOPE.value: telescope,
            ParamsKey.REINGEST.value: False,
            ParamsKey.NGAS_INGEST.value: False,
            ParamsKey.CALIBRATE.value: False,
            ParamsKey.INGESTION_PATH.value: "/home/mchammer/evla/parallel-prod",
        }
    }
    param_values_dict = params_dict[ParamsKey.PARAMETERS.value]

    params = ManifestParameters(
        telescope=param_values_dict[ParamsKey.TELESCOPE.value],
        reingest=param_values_dict[ParamsKey.REINGEST.value],
        ngas_ingest=param_values_dict[ParamsKey.NGAS_INGEST.value],
        calibrate=param_values_dict[ParamsKey.CALIBRATE.value],
        staging_source_dir=param_values_dict[ParamsKey.INGESTION_PATH.value],
    )

    params_json = params.to_json()
    for key, val in params_json.items():
        assert (
            val == params_dict[key] if isinstance(params_dict[key], bool) else str(params_dict[key])
        )


@pytest.mark.skip("TODO")
def test_params_properly_formatted():
    """
    TODO
    :return:
    """
    raise NotImplementedError


def test_input_sp_well_formed():
    """
    Make sure our InputScienceProduct makes nice JSON

    :return:
    """
    locator = "uid://evla/calibration/vanilla_heath_bar_crunch_1a23e"
    # single science product
    sp_dict = {
        "type": ScienceProductType.EVLA_CAL.value,
        "locator": locator,
    }

    sp_in = InputScienceProduct(sp_type=ScienceProductType.EVLA_CAL.value, locator=locator)
    assert sp_in.to_json() == sp_dict


def test_input_group_well_formed():
    """
    Make sure our InputGroup makes nice JSON

    :return:
    """
    sp1 = InputScienceProduct(
        sp_type=ScienceProductType.EXEC_BLOCK.value,
        locator="uid://evla/execblock/coffee_heath_bar_crunch_7a23f",
    )
    sp1_json = sp1.to_json()

    sp2 = InputScienceProduct(
        sp_type=ScienceProductType.EVLA_CAL.value, locator="uid://evla/execblock/mint_oreo_omg_omg"
    )
    sp2_json = sp2.to_json()

    expected = {
        IngestionManifestKey.INPUT_GROUP.value: {
            IngestionManifestKey.SCIENCE_PRODUCTS.value: [sp1_json, sp2_json]
        }
    }

    ingroup = InputGroup(science_products=[sp1, sp2])
    actual = ingroup.to_json()
    assert actual.keys() == expected.keys()

    actual = actual[IngestionManifestKey.INPUT_GROUP.value]
    expected = expected[IngestionManifestKey.INPUT_GROUP.value]
    assert actual.keys() == expected.keys()

    actual = actual[IngestionManifestKey.SCIENCE_PRODUCTS.value]
    expected = expected[IngestionManifestKey.SCIENCE_PRODUCTS.value]
    assert len(actual) == len(expected) == 2
    marvin = actual[0]
    trillian = expected[0]
    for key, _ in marvin.items():
        assert trillian[key] == marvin[key]

    marvin = actual[1]
    trillian = expected[1]
    for key, _ in marvin.items():
        assert trillian[key] == marvin[key]


def test_ancillary_product_well_formed():
    """
    The JSON shouldn't contain empty fields

    :return:
    """
    ap1 = AncillaryProduct(type=AncillaryProductType.LOG, filename="without_feathers.tar")
    expected = {"type": AncillaryProductType.LOG.value, "filename": "without_feathers.tar"}
    actual = ap1.to_json()

    assert actual == expected


def test_output_group_well_formed():
    """
    Make sure our OutputScienceProduct makes nice JSON

    :return:
    """
    sp1 = OutputScienceProduct(type=ScienceProductType.IMAGE, filename="see_no_evil.fits")
    sp2 = OutputScienceProduct(type=ScienceProductType.IMAGE, filename="hear_no_evil.fits")
    ap1 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="without_feathers.tar"
    )
    ap2 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="with_feathers.tar"
    )
    opg = OutputGroup(science_products=[sp1, sp2], ancillary_products=[ap1, ap2])
    opg_json = opg.to_json()
    dumped = json.dumps(opg_json)

    assert (
        dumped == '{"output_group": '
        '{"science_products": '
        '[{"type": "image", "filename": "see_no_evil.fits"}, '
        '{"type": "image", "filename": "hear_no_evil.fits"}], '
        '"ancillary_products": [{"type": "pipeline_artifacts", '
        '"filename": "without_feathers.tar"}, '
        '{"type": "pipeline_artifacts", "filename": "with_feathers.tar"}]}}'
    )


@pytest.mark.skip("TODO, or maybe not")
def test_input_group_properly_formatted():
    """
    Does the InputGroup get written to the file in the expected format?

    :return:
    """
    sp1 = InputScienceProduct(
        sp_type=ScienceProductType.EXEC_BLOCK.value,
        locator="uid://evla/execblock/coffee_heath_bar_crunch_7a23f",
    )

    ingroup = InputGroup(science_products=[sp1])
    ig_dict = json.loads(ingroup.to_json())
    ig_text = json.dumps(ig_dict, indent=4)
    expected = """
  "input-group": {
    "science_products": [
      {
        "type": "calibration",
        "locator": "uid://evla/execblock/coffee_heath_bar_crunch_7a23f"
      }
    ]
  },
    """
    assert ig_text == expected


@pytest.mark.skip("TODO")
def test_output_group_properly_formatted():
    """
    TODO
    Does the OutputGroup get written to the file in the expected format?

    :return:
    """
    raise NotImplementedError


def test_builds_cal_manifest_as_expected(ingest_path: Path):
    """
    When we create an EVLA calibration ingestion manifest, does it contain all it should?
    We'll make a manifest that should look like our example and make sure it does.

    :return:
    """

    populate_ingest_path_for_manifest_evla_cal_example(ingest_path)

    locator = "uid://evla/execblock/fjdsakljfkdlsajfkldsa"
    IngestionManifestBuilder(
        telescope=Telescope.EVLA,
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.EVLA_CAL,
        locator=locator,
    ).build()

    manifest_file = find_manifest(ingest_path)
    with open(manifest_file, "r") as mf_in:
        manifest_content = dict(json.load(mf_in).items())

    # check parameters
    parameters = manifest_content["parameters"]["parameters"]
    for param in ["reingest", "ngas_ingest", "calibrate"]:
        assert parameters[param] is False
    assert parameters[ParamsKey.INGESTION_PATH.value] == str(ingest_path)

    # check input group
    input_group = manifest_content[IngestionManifestKey.INPUT_GROUP.value][
        IngestionManifestKey.INPUT_GROUP.value
    ]
    assert len(input_group[IngestionManifestKey.SCIENCE_PRODUCTS.value]) == 1
    science_product = input_group[IngestionManifestKey.SCIENCE_PRODUCTS.value][0]
    assert science_product["locator"] == locator

    # check output group
    output_group = manifest_content[IngestionManifestKey.OUTPUT_GROUP.value][
        IngestionManifestKey.OUTPUT_GROUP.value
    ]
    science_products = output_group[IngestionManifestKey.SCIENCE_PRODUCTS.value]
    assert len(science_products) == 1
    ancillary_products = output_group[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
    assert len(ancillary_products) == 2

    shutil.rmtree(ingest_path)


def populate_ingest_path_for_manifest_evla_cal_example(ingestion_path: Path):
    """
    Create fake input files to match EVLA CAL manifest example

    :param ingestion_path:
    :return:
    """
    weblog_file = ingestion_path / "qrs.weblog.tgz"
    weblog_file.touch()
    cal_file = ingestion_path / "XYZ-abc+TMN.O00.tar"
    cal_file.touch()
