import logging
from typing import Dict

import requests
from aat_wrest.observation_wrester import WrestObservationMetadata

from aat_wrest.utilities import MDDBConnector
from messaging.router import Router, on_message
from pycapo import CapoConfig
from requests import Response

from workspaces.products.services.interfaces import ArchiveServiceIF

logger = logging.getLogger(__name__)


class ArchiveService(ArchiveServiceIF):
    """
    The archive service: clients access this to interface with the archive services
    """

    def __init__(self):
        self.message_router = Router("archive")
        self.message_router.register(self)

    @on_message(message="Ingestion complete")
    def on_ingestion_complete(self, **message: Dict):
        """
        Accept a ingestion-complete message from the archive

        :param message: the ingestion-complete message
        """

        logdata = message["logData"]

        # log message for Demo
        logger.info(f"ingestion-complete event: {logdata}")

        execblock_id = logdata["execblock_id"]
        sdm_id = logdata["fileset_id"]
        project_code = logdata["project_code"]

        wrester = WrestObservationMetadata(MDDBConnector(), sdm_id=sdm_id)
        metadata = wrester.wrest_observation_info()
        metadata["sdm_id"] = sdm_id
        spl = metadata["spl"]
        del metadata["spl"]

        response = self._create_std_calibration_request(spl, metadata)

        logger.debug(
            "Got result %s with type %s and body %s",
            response,
            type(response),
            response.content,
        )
        return spl, execblock_id, sdm_id, project_code

    @staticmethod
    def _create_std_calibration_request(spl: str, metadata: Dict[str, str]) -> Response:
        """
        Creates a new standard calibration with information given from an ingestion-completed event

        :param spl: Science product locator of observation
        :param metadata: Dict with the following information inside:
            - sdm_id: SDM ID (sometimes called fileset id)
            - bands: Radio bands for observation
            - array_config: Configuration of array telescopes
            - obs_start_time: Start time of observation
            - obs_end_time: End time of observation
        :return: HTTP Response for attempt to create std_calibration request
        """
        capability_url = (
            CapoConfig().settings("edu.nrao.archive.workspaces.CapabilitySettings").serviceUrl
        )
        analyst_email = (
            CapoConfig().settings("edu.nrao.archive.workspaces.NotificationSettings").analystEmail
        )
        parameters = {
            "parameters": {
                "product_locator": spl,
                "user_email": analyst_email,
                "metadata": metadata,
            }
        }

        return requests.post(
            f"{capability_url}/capability/std_calibration/request/create", json=parameters
        )
